package eiffel

import (
	"context"
	"encoding/base64"
	"encoding/json"
	"errors"
	"fmt"
	"log"
	"net/http"
	"os"
	"strings"
	"time"

	"github.com/google/uuid"
	"github.com/org-harmony/harmony/src/app/template"
	"github.com/org-harmony/harmony/src/app/template/parser"
	"github.com/org-harmony/harmony/src/app/user"
	"github.com/org-harmony/harmony/src/core/config"
	"github.com/org-harmony/harmony/src/core/event"
	"github.com/org-harmony/harmony/src/core/hctx"
	"github.com/org-harmony/harmony/src/core/persistence"
	"github.com/org-harmony/harmony/src/core/util"
	"github.com/org-harmony/harmony/src/core/web"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"
)

const (
	// TemplateDisplayString will display the rule value as text.
	TemplateDisplayString TemplateDisplayType = "text"
	// TemplateDisplayInputTypeText will display the rule value as a text input.
	TemplateDisplayInputTypeText TemplateDisplayType = "input-text"
	// TemplateDisplayInputTypeTextarea will display the rule value as a textarea.
	TemplateDisplayInputTypeTextarea TemplateDisplayType = "input-textarea"
	// TemplateDisplayInputTypeSingleSelect will display the rule value as an input field with datalist and single select.
	TemplateDisplayInputTypeSingleSelect TemplateDisplayType = "input-single-select"
)

var (
	// ErrTemplateNotFound will be displayed to the user if the template could not be found.
	ErrTemplateNotFound = errors.New("eiffel.elicitation.template.not-found")
	// ErrTemplateVariantNotFound will be displayed to the user if the template variant could not be found.
	ErrTemplateVariantNotFound = errors.New("eiffel.elicitation.template.variant.not-found")
)

// TemplateDisplayType specifies how a rule should be displayed in the UI.
type TemplateDisplayType string

// TemplateFormData is the data that is passed to the template rendering the elicitation form.
type TemplateFormData struct {
	Template *BasicTemplate
	// Variant is the currently selected variant. This might be through a specified variant name parameter
	// or as a default value because no variant was explicitly specified. However, Variant is expected to be filled.
	Variant *BasicVariant
	// VariantKey is the key through which the variant was selected. It is not the name of the variant.
	// If the variant was auto-selected as default the key will still be filled.
	VariantKey string
	// DisplayTypes is a map of rule names to display types. The rule names are the keys of the BasicTemplate.Rules map.
	// The display types are used to determine how the rule should be displayed in the UI.
	DisplayTypes map[string]TemplateDisplayType
	// TemplateID is the ID of the template that is currently being rendered.
	TemplateID uuid.UUID
	// CopyAfterParse is a flag indicating if the user wants to copy the parsed requirement to the clipboard.
	CopyAfterParse bool
	// ParsingResult is the result of the parsing process. This can be empty if no parsing was done yet.
	ParsingResult *parser.ParsingResult
	// SegmentMap is a map of rule names to their corresponding segment value.
	// This is used to fill the segments with their values after parsing.
	SegmentMap map[string]string
	// NeglectOptional is a flag indicating if optional rules (inputs) should be displayed different from non-optional rules.
	NeglectOptional bool
	ProjectID       string
	EditMode        bool
}

// SearchTemplateData contains templates to render as search results and a flag indicating if the query was too short.
type SearchTemplateData struct {
	Templates []*template.Template
	// Deprecated this is expected to be unnecessary with the current implementation of EIFFEL
	QueryTooShort bool
}

type HTMXTriggerParsingSuccessEvent struct {
	ParsingSuccessEvent *parser.ParsingResult `json:"parsingSuccessEvent"`
}

func ExportToJSON(filename string) error {
	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
	defer cancel()

	log.Println("Fetching data from MongoDB...")
	cursor, err := mongoCollection.Find(ctx, bson.M{})
	if err != nil {
		log.Printf("Failed to fetch data: %v", err)
		return fmt.Errorf("failed to fetch data from MongoDB: %w", err)
	}
	defer cursor.Close(ctx)

	var requirements []bson.M
	if err := cursor.All(ctx, &requirements); err != nil {
		log.Printf("Failed to parse data: %v", err)
		return fmt.Errorf("failed to parse MongoDB data: %w", err)
	}
	log.Printf("Fetched %d documents from MongoDB", len(requirements))

	log.Println("Creating JSON file...")
	file, err := os.Create(filename)
	if err != nil {
		log.Printf("Failed to create file: %v", err)
		return fmt.Errorf("failed to create JSON file: %w", err)
	}
	defer file.Close()

	log.Println("Writing data to JSON file...")
	if err := json.NewEncoder(file).Encode(requirements); err != nil {
		log.Printf("Failed to write JSON data: %v", err)
		return fmt.Errorf("failed to write JSON data to file: %w", err)
	}

	log.Printf("Exported data to JSON file: %s", filename)
	return nil
}

func exportJSONHandler(cfg Cfg) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		filename := "requirements.json"

		if err := ExportToJSON(filename); err != nil {
			log.Printf("Error exporting to JSON: %v", err)
			http.Error(w, fmt.Sprintf("Failed to export JSON: %v", err), http.StatusInternalServerError)
			return
		}

		w.Header().Set("Content-Disposition", fmt.Sprintf("attachment; filename=%s", filename))
		w.Header().Set("Content-Type", "application/json")
		http.ServeFile(w, r, filename)
	})
}

func ExportToTXT(filename string) error {
	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
	defer cancel()

	log.Println("Fetching data from MongoDB...")
	cursor, err := mongoCollection.Find(ctx, bson.M{})
	if err != nil {
		log.Printf("Failed to fetch data: %v", err)
		return fmt.Errorf("failed to fetch data from MongoDB: %w", err)
	}
	defer cursor.Close(ctx)

	var requirements []bson.M
	if err := cursor.All(ctx, &requirements); err != nil {
		log.Printf("Failed to parse data: %v", err)
		return fmt.Errorf("failed to parse MongoDB data: %w", err)
	}
	log.Printf("Fetched %d documents from MongoDB", len(requirements))

	log.Println("Creating TXT file...")
	file, err := os.Create(filename)
	if err != nil {
		log.Printf("Failed to create file: %v", err)
		return fmt.Errorf("failed to create TXT file: %w", err)
	}
	defer file.Close()

	log.Println("Writing data to TXT file...")
	for _, requirement := range requirements {

		line, err := bson.MarshalExtJSON(requirement, true, true)
		if err != nil {
			log.Printf("Failed to marshal document to text: %v", err)
			return fmt.Errorf("failed to convert document to text: %w", err)
		}
		_, err = file.WriteString(string(line) + "\n")
		if err != nil {
			log.Printf("Failed to write line to TXT file: %v", err)
			return fmt.Errorf("failed to write line to TXT file: %w", err)
		}
	}

	log.Printf("Exported data to TXT file: %s", filename)
	return nil
}

func exportTXTHandler(cfg Cfg) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		filename := "requirements.txt"

		if err := ExportToTXT(filename); err != nil {
			log.Printf("Error exporting to TXT: %v", err)
			http.Error(w, fmt.Sprintf("Failed to export TXT: %v", err), http.StatusInternalServerError)
			return
		}

		w.Header().Set("Content-Disposition", fmt.Sprintf("attachment; filename=%s", filename))
		w.Header().Set("Content-Type", "text/plain")
		http.ServeFile(w, r, filename)
	})
}

func deleteProjectHandler(appCtx *hctx.AppCtx, webCtx *web.Ctx) http.Handler {
	return web.NewController(appCtx, webCtx, func(io web.IO) error {
		id := web.URLParam(io.Request(), "id")
		ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
		defer cancel()

		var filter bson.M
		if len(id) == 24 {
			if objectID, err := primitive.ObjectIDFromHex(id); err == nil {
				filter = bson.M{"_id": objectID}
			} else {
				filter = bson.M{"_id": id}
			}
		} else {
			filter = bson.M{"_id": id}
		}

		if _, err := mongoCollection.DeleteMany(ctx, bson.M{"project_id": id}); err != nil {
			return fmt.Errorf("Fehler beim Löschen der zugehörigen Requirements: %w", err)
		}

		if _, err := mongoCollection.DeleteOne(ctx, filter); err != nil {
			return fmt.Errorf("Fehler beim Löschen des Projekts: %w", err)
		}

		io.Response().Header().Set("Location", "/eiffel/projects")
		io.Response().WriteHeader(http.StatusSeeOther) // 303 Redirect
		return nil
	})
}

func deleteRequirementHandler(appCtx *hctx.AppCtx, webCtx *web.Ctx) http.Handler {
	return web.NewController(appCtx, webCtx, func(io web.IO) error {
		id := web.URLParam(io.Request(), "id")
		id = strings.TrimPrefix(id, "ObjectID(\"")
		id = strings.TrimSuffix(id, "\")")

		if len(id) != 24 {
			return fmt.Errorf("ungültige ID-Länge: %s", id)
		}

		objectID, err := primitive.ObjectIDFromHex(id)
		if err != nil {
			return fmt.Errorf("ungültige ID: %w", err)
		}

		ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
		defer cancel()

		if _, err := mongoCollection.DeleteOne(ctx, bson.M{"_id": objectID}); err != nil {
			return fmt.Errorf("Fehler beim Löschen des Requirements: %w", err)
		}

		io.Response().Header().Set("Location", "/eiffel/projects")
		io.Response().WriteHeader(http.StatusSeeOther) // 303 Redirect
		return nil
	})
}

func updateRequirementHandler(appCtx *hctx.AppCtx, webCtx *web.Ctx) http.Handler {
	return web.NewController(appCtx, webCtx, func(io web.IO) error {
		id := web.URLParam(io.Request(), "id")

		// Bereinige die ID
		id = strings.TrimPrefix(id, "ObjectID(\"")
		id = strings.TrimSuffix(id, "\")")

		if len(id) != 24 {
			return fmt.Errorf("ungültige ID: %s", id)
		}

		objectID, err := primitive.ObjectIDFromHex(id)
		if err != nil {
			return fmt.Errorf("ungültige ID: %w", err)
		}

		// Neue Daten aus dem Formular abrufen
		var formData struct {
			Requirement string `json:"requirement"`
		}
		if err := io.Request().ParseForm(); err != nil {
			return fmt.Errorf("Fehler beim Parsen des Formulars: %w", err)
		}
		formData.Requirement = io.Request().FormValue("requirement")

		// Requirement in der Datenbank aktualisieren
		ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
		defer cancel()

		_, err = mongoCollection.UpdateOne(
			ctx,
			bson.M{"_id": objectID},
			bson.M{"$set": bson.M{"parsing_result.requirement": formData.Requirement}},
		)
		if err != nil {
			return fmt.Errorf("Fehler beim Aktualisieren des Requirements: %w", err)
		}

		io.Response().Header().Set("Location", "/eiffel/projects") // Zurück zur Projektseite
		io.Response().WriteHeader(http.StatusSeeOther)
		return nil
	})
}

// RegisterController registers the controllers as well as the navigation and event listeners.
func RegisterController(appCtx *hctx.AppCtx, webCtx *web.Ctx) {
	cfg := Cfg{}
	util.Ok(config.C(&cfg, config.From("eiffel"), config.Validate(appCtx.Validator)))

	// TODO move this to module init when module manager is implemented (see subscribeEvents)
	subscribeEvents(appCtx)

	registerNavigation(appCtx, webCtx)

	router := webCtx.Router.With(user.LoggedInMiddleware(appCtx))

	router.Get("/eiffel", eiffelElicitationPage(cfg, appCtx, webCtx).ServeHTTP)
	router.Get("/eiffel/{templateID}", eiffelElicitationPage(cfg, appCtx, webCtx).ServeHTTP)
	router.Get("/eiffel/{templateID}/{variant}", eiffelElicitationPage(cfg, appCtx, webCtx).ServeHTTP)
	router.Post("/eiffel/{templateID}/{variant}", eiffelElicitationPage(cfg, appCtx, webCtx).ServeHTTP)

	router.Get("/eiffel/elicitation/templates/search/modal", searchModal(appCtx, webCtx).ServeHTTP)
	router.Post("/eiffel/elicitation/templates/search", searchTemplate(appCtx, webCtx).ServeHTTP)
	router.Get("/eiffel/elicitation/{templateID}", elicitationTemplate(cfg, appCtx, webCtx, true).ServeHTTP)
	router.Get("/eiffel/elicitation/{templateID}/{variant}", elicitationTemplate(cfg, appCtx, webCtx, false).ServeHTTP)
	router.Post("/eiffel/elicitation/{templateID}/{variant}", parseRequirement(cfg, appCtx, webCtx).ServeHTTP)
	//Export Endpunkte
	router.Get("/eiffel/export/json", exportJSONHandler(cfg).ServeHTTP)
	router.Get("/eiffel/export/txt", exportTXTHandler(cfg).ServeHTTP)
	// TO DO CRUD-Endpunkte müssen noch überarbeitet werden
	router.Post("/eiffel/requirements/{id}/update", updateRequirementHandler(appCtx, webCtx).ServeHTTP)

	router.Post("/eiffel/projects/{id}/delete", deleteProjectHandler(appCtx, webCtx).ServeHTTP)
	router.Post("/eiffel/requirements/{id}/delete", deleteRequirementHandler(appCtx, webCtx).ServeHTTP)
	//Projekt Endpunkt
	router.Get("/eiffel/projects", getProjectPackagesHandler(appCtx, webCtx).ServeHTTP)
}

func subscribeEvents(appCtx *hctx.AppCtx) {
	// TODO remove this with module manager
	appCtx.EventManager.Subscribe("template.config.validate", func(event event.Event, args *event.PublishArgs) error {
		validateEvent, ok := event.Payload().(*template.ValidateTemplateConfigEvent)
		if !ok {
			return nil
		}
		if strings.ToLower(validateEvent.TemplateType) != BasicTemplateType {
			return nil
		}
		if validateEvent.DidValidate {
			return nil
		}
		validateEvent.DidValidate = true

		ebt := &BasicTemplate{}
		// Important notice: Unmarshalling is always case-insensitive if no other match could be found.
		// Therefore, NAME will be unmarshalled to Name. Keep this in mind.
		err := json.Unmarshal([]byte(validateEvent.Config), ebt)
		if err != nil {
			return err
		}

		validationErrs := ebt.Validate(appCtx.Validator, RuleParsers())
		if len(validationErrs) > 0 {
			validateEvent.AddErrors(validationErrs...)
			return nil
		}

		return nil
	}, event.DefaultPriority)
}

func registerNavigation(appCtx *hctx.AppCtx, webCtx *web.Ctx) {
	webCtx.Navigation.Add("eiffel.elicitation", web.NavItem{
		URL:  "/eiffel",
		Name: "harmony.menu.eiffel",
		Display: func(io web.IO) (bool, error) {
			return true, nil
		},
		Position: 100,
	})

	// Projekt-Sammlung Button
	webCtx.Navigation.Add("eiffel.project.collection", web.NavItem{
		URL:  "/eiffel/projects", // Die URL zu deiner Projekt-Sammlung-Seite
		Name: "Projekt-Sammlung", // Der Name, der angezeigt wird
		Display: func(io web.IO) (bool, error) {
			return true, nil
		},
		Position: 110, // Position in der Navigation; erscheint direkt nach dem EIFFEL-Button
	})
}

type ProjectCollectionData struct {
	Projects []bson.M
}

func getProjectPackagesHandler(appCtx *hctx.AppCtx, webCtx *web.Ctx) http.Handler {
	return web.NewController(appCtx, webCtx, func(io web.IO) error {
		ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
		defer cancel()

		// MongoDB Aggregation: Gruppieren nach project_id
		pipeline := mongo.Pipeline{
			bson.D{{
				"$group", bson.D{
					{"_id", "$project_id"},
					{"requirements", bson.D{{"$push", "$$ROOT"}}},
				},
			}},
		}

		cursor, err := mongoCollection.Aggregate(ctx, pipeline)
		if err != nil {
			log.Printf("Error fetching project packages: %v", err)
			return io.InlineError(web.ErrInternal, err)
		}
		defer cursor.Close(ctx)

		var results []bson.M
		if err := cursor.All(ctx, &results); err != nil {
			log.Printf("Error parsing project packages: %v", err)
			return io.InlineError(web.ErrInternal, err)
		}

		// Verpacke die Ergebnisse in die Struktur
		data := ProjectCollectionData{
			Projects: results,
		}

		//log.Printf("Fetched projects: %+v", data)

		// Erfolgsnachricht für das Template vorbereiten
		successMessage := []string{"Projects fetched successfully"}

		// Template rendern und Daten korrekt übergeben
		return io.Render(
			web.NewFormData(data, successMessage, err),
			"project-collection",
			"eiffel/_project-collection.go.html",
		)
	})
}

func eiffelElicitationPage(cfg Cfg, appCtx *hctx.AppCtx, webCtx *web.Ctx) http.Handler {
	templateRepository := util.UnwrapType[template.Repository](appCtx.Repository(template.RepositoryName))
	sessionStore := util.UnwrapType[user.SessionRepository](appCtx.Repository(user.SessionRepositoryName))

	return web.NewController(appCtx, webCtx, func(io web.IO) error {
		templateID := web.URLParam(io.Request(), "templateID")
		variantKey := web.URLParam(io.Request(), "variant")
		requirementID := io.Request().PostFormValue("requirementID") // POST-Wert abrufen
		editMode := io.Request().PostFormValue("edit") == "true"
		// Prüfe, ob `templateID` leer ist, und rendere ein leeres Formular
		if templateID == "" {
			log.Println("templateID ist leer, rendere leeres Formular.")
			return renderElicitationPage(io, TemplateFormData{NeglectOptional: cfg.NeglectOptional}, nil, nil)
		}
		// RequirementID bereinigen
		requirementID = strings.TrimPrefix(requirementID, "ObjectID(\"")
		requirementID = strings.TrimSuffix(requirementID, "\")")

		// Debugging: Log requirementID und editMode
		log.Printf("Received requirementID: %s, editMode: %t", requirementID, editMode)

		// Grundlegende Formulardaten laden
		formData, err := TemplateFormFromRequest(
			io.Context(),
			templateID,
			variantKey,
			templateRepository,
			RuleParsers(),
			appCtx.Validator,
			true,
		)

		formData.NeglectOptional = cfg.NeglectOptional
		formData.CopyAfterParse = CopyAfterParseSetting(io.Request(), sessionStore, true)

		// Wenn "Edit-Modus" aktiv ist, lade Segment-Daten aus der Datenbank
		if editMode && requirementID != "" {
			ctx := io.Context()

			// Validate requirementID as ObjectID
			filter := bson.M{}
			if len(requirementID) == 24 {
				objectID, parseErr := primitive.ObjectIDFromHex(requirementID)
				if parseErr == nil {
					filter["_id"] = objectID
				} else {
					log.Printf("Invalid ObjectID format: %v", parseErr)
					return fmt.Errorf("Invalid Requirement ID: %s", requirementID)
				}
			} else {
				log.Printf("RequirementID is not a valid ObjectID: %s", requirementID)
				return fmt.Errorf("RequirementID is not valid: %s", requirementID)
			}

			// Query database for requirement
			var requirement bson.M
			dbErr := mongoCollection.FindOne(ctx, filter).Decode(&requirement)
			if dbErr != nil {
				log.Printf("Requirement not found: %v", dbErr)
				return fmt.Errorf("Requirement not found: %w", dbErr)
			}

			log.Printf("Fetched Requirement: %+v", requirement)

			// Extract SegmentMap from the database result
			segmentMap := make(map[string]string)
			if rawSegmentMap, ok := requirement["segment_map"].(bson.M); ok {
				for key, value := range rawSegmentMap {
					if strValue, ok := value.(string); ok {
						segmentMap[key] = strValue
					}
				}
			}

			if projectIDValue, ok := requirement["project_id"].(string); ok {
				formData.ProjectID = projectIDValue
				log.Printf("Loaded Project ID: %s", formData.ProjectID)
			} else {
				log.Printf("Project ID not found in requirement")
			}
			log.Printf("Loaded SegmentMap: %+v", segmentMap)

			// Assign SegmentMap to formData
			formData.SegmentMap = segmentMap
			formData.EditMode = editMode
			//Lösche das Requirement wieder
			if _, err := mongoCollection.DeleteMany(ctx, bson.M{"project_id": requirementID}); err != nil {
				return fmt.Errorf("Fehler beim Löschen der zugehörigen Requirements: %w", err)
			}
			if _, err := mongoCollection.DeleteOne(ctx, filter); err != nil {
				return fmt.Errorf("Fehler beim Löschen des Requirements: %w", err)
			}
		}

		// Debugging: Log the final FormData before rendering
		log.Printf("Final FormData: %+v", formData)
		log.Printf("FormData: %+v", formData)
		// Render the form
		return renderElicitationPage(io, formData, nil, []error{err})
	})
}

func renderElicitationPage(io web.IO, data TemplateFormData, success []string, errs []error) error {
	log.Printf("Render context: %+v", web.NewFormData(data, success, errs...))

	return io.Render(
		web.NewFormData(data, success, errs...),
		"eiffel.elicitation.page",
		"eiffel/elicitation-page.go.html",
		"eiffel/_elicitation-template.go.html",
		"eiffel/_form-elicitation.go.html",
		"eiffel/_list-requirements.go.html",
	)
}

func searchModal(appCtx *hctx.AppCtx, webCtx *web.Ctx) http.Handler {
	templateRepository := util.UnwrapType[template.Repository](appCtx.Repository(template.RepositoryName))

	return web.NewController(appCtx, webCtx, func(io web.IO) error {
		ctx := io.Context()
		templates, err := templateRepository.FindByQueryForTypeAndUser(ctx, "", BasicTemplateType, user.MustCtxUser(ctx))
		if err != nil && !errors.Is(err, persistence.ErrNotFound) {
			return io.InlineError(web.ErrInternal, err)
		}

		return io.Render(
			&SearchTemplateData{Templates: templates},
			"eiffel.template.search.modal",
			"eiffel/_modal-template-search.go.html",
			"eiffel/_template-search-result.go.html",
		)
	})
}

func searchTemplate(appCtx *hctx.AppCtx, webCtx *web.Ctx) http.Handler {
	templateRepository := util.UnwrapType[template.Repository](appCtx.Repository(template.RepositoryName))

	return web.NewController(appCtx, webCtx, func(io web.IO) error {
		request := io.Request()
		err := request.ParseForm()
		if err != nil {
			return io.InlineError(web.ErrInternal, err)
		}

		query := request.FormValue("search")
		// query too short was removed as it is expected to be unnecessary

		ctx := io.Context()
		templates, err := templateRepository.FindByQueryForTypeAndUser(ctx, query, BasicTemplateType, user.MustCtxUser(ctx))
		if err != nil && !errors.Is(err, persistence.ErrNotFound) {
			return io.InlineError(web.ErrInternal, err)
		}

		return io.Render(
			&SearchTemplateData{Templates: templates},
			"eiffel.template.search.result",
			"eiffel/_template-search-result.go.html",
		)
	})
}

func elicitationTemplate(cfg Cfg, appCtx *hctx.AppCtx, webCtx *web.Ctx, defaultFirstVariant bool) http.Handler {
	templateRepository := util.UnwrapType[template.Repository](appCtx.Repository(template.RepositoryName))
	sessionStore := util.UnwrapType[user.SessionRepository](appCtx.Repository(user.SessionRepositoryName))

	return web.NewController(appCtx, webCtx, func(io web.IO) error {
		templateID := web.URLParam(io.Request(), "templateID")
		variant := web.URLParam(io.Request(), "variant")

		io.Response().Header().Set("HX-Push-URL", fmt.Sprintf("/eiffel/%s", templateID))

		formData, err := TemplateFormFromRequest(
			io.Context(),
			templateID,
			variant,
			templateRepository,
			RuleParsers(),
			appCtx.Validator,
			defaultFirstVariant,
		)
		if err != nil {
			return io.InlineError(err)
		}

		formData.NeglectOptional = cfg.NeglectOptional
		formData.CopyAfterParse = CopyAfterParseSetting(io.Request(), sessionStore, true)

		io.Response().Header().Set("HX-Push-URL", fmt.Sprintf("/eiffel/%s/%s", templateID, formData.VariantKey))

		return io.Render(
			web.NewFormData(formData, nil),
			"eiffel.elicitation.template",
			"eiffel/_elicitation-template.go.html",
			"eiffel/_form-elicitation.go.html",
		)
	})
}

var mongoClient *mongo.Client
var mongoCollection *mongo.Collection

func InitMongoDB() error {
	clientOptions := options.Client().ApplyURI("mongodb://localhost:27017")
	client, err := mongo.Connect(context.TODO(), clientOptions)
	if err != nil {
		return fmt.Errorf("failed to connect to MongoDB: %w", err)
	}

	err = client.Ping(context.TODO(), nil)
	if err != nil {
		return fmt.Errorf("failed to ping MongoDB: %w", err)
	}

	mongoCollection = client.Database("harmony").Collection("requirements")
	log.Println("Connected to MongoDB")
	return nil
}

func saveToMongoDB(formData *TemplateFormData) error {
	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	document := bson.M{
		"template_id":    formData.TemplateID.String(),
		"variant_key":    formData.VariantKey,
		"segment_map":    formData.SegmentMap,
		"parsing_result": formData.ParsingResult,
		"project_id":     formData.ProjectID,
		"created_at":     time.Now(),
	}

	_, err := mongoCollection.InsertOne(ctx, document)
	return err
}

func parseRequirement(cfg Cfg, appCtx *hctx.AppCtx, webCtx *web.Ctx) http.Handler {
	templateRepository := util.UnwrapType[template.Repository](appCtx.Repository(template.RepositoryName))
	sessionStore := util.UnwrapType[user.SessionRepository](appCtx.Repository(user.SessionRepositoryName))

	return web.NewController(appCtx, webCtx, func(io web.IO) error {
		log.Println("parseRequirement called")
		request := io.Request()
		ctx := request.Context()
		parsers := RuleParsers()

		// **Formulardaten explizit parsen**
		if err := request.ParseForm(); err != nil {
			log.Printf("Error parsing form data: %v", err)
			return io.InlineError(web.ErrInternal, err)
		}

		// **Action prüfen: validate oder save**
		action := request.FormValue("action")
		log.Printf("Form Action: %s", action)

		// **ProjectID aus Formulardaten extrahieren**
		projectID := request.FormValue("project_id")
		log.Printf("Captured Project ID: %s", projectID)

		templateID := web.URLParam(request, "templateID")
		variant := web.URLParam(request, "variant")

		// Template-Daten lesen
		formData, err := TemplateFormFromRequest(
			ctx,
			templateID,
			variant,
			templateRepository,
			parsers,
			appCtx.Validator,
			false,
		)
		if err != nil {
			return io.InlineError(err)
		}
		formData.ProjectID = projectID

		segmentMap, err := SegmentMapFromRequest(request, len(formData.Variant.Rules))
		if err != nil {
			return io.InlineError(web.ErrInternal, err)
		}
		formData.SegmentMap = segmentMap

		// **Anforderung prüfen (validate)**
		parsingResult, err := formData.Template.Parse(ctx, parsers, formData.VariantKey, SegmentMapToSegments(segmentMap)...)
		formData.ParsingResult = &parsingResult
		log.Printf("Captured formData: %+v\n", formData)

		var s []string
		if parsingResult.Flawless() {
			s = []string{"eiffel.elicitation.parse.flawless-success"}
		} else if parsingResult.Ok() {
			s = []string{"eiffel.elicitation.parse.success"}
		}

		// **HX-Trigger wird für Validierungsereignisse genutzt**
		if parsingResult.Ok() {
			triggerEvent := &HTMXTriggerParsingSuccessEvent{ParsingSuccessEvent: &parsingResult}
			triggerEventJSON, err := json.Marshal(triggerEvent)
			if err != nil {
				return io.InlineError(web.ErrInternal, err)
			}

			// Using HX-Trigger header here is not possible because we could potentially use unicode characters in our response.
			// To escape them we have to use base64 encoding and a custom header.
			io.Response().Header().Set("ParsingSuccessEvent", base64.URLEncoding.EncodeToString(triggerEventJSON))
		}

		// **Speichern nur bei Aktion "save"**
		if action == "save" {
			log.Println("Saving requirement to MongoDB...")
			err = saveToMongoDB(&formData)
			if err != nil {
				log.Printf("Failed to save formData to MongoDB: %v\n", err)
				return io.InlineError(web.ErrInternal, err)
			}
			log.Println("Form data successfully saved to MongoDB")
			s = append(s, "eiffel.elicitation.save.success")
		} else {
			log.Println("Validation complete, no data saved to MongoDB")
		}

		formData.NeglectOptional = cfg.NeglectOptional
		formData.CopyAfterParse = CopyAfterParseSetting(request, sessionStore, false)

		return io.Render(web.NewFormData(formData, s, err), "eiffel.elicitation.form", "eiffel/_form-elicitation.go.html")
	})
}
